import pytest

from netqasm.util.string import group_by_word, is_float, is_number, is_variable_name


@pytest.mark.parametrize(
    "string, seperator, brackets, expected",
    [
        ("hello", " ", "()", ["hello"]),
        ("hello world", " ", "()", ["hello", "world"]),
        ("hello world", "o", "()", ["hell", " w", "rld"]),
        ("(hello world)", " ", "()", ["(hello world)"]),
        ("(hello world)", " ", "[]", ["(hello", "world)"]),
        ("hello (world)", " ", "()", ["hello", "(world)"]),
        ("hello (wo(rld)", " ", "()", ["hello", "(wo(rld)"]),
        ("hello (wo)rld)", " ", "()", ["hello", "(wo)rld)"]),
        ("hello (wo rld)", " ", "()", ["hello", "(wo rld)"]),
        ("hello wo rld", " ", "()", ["hello", "wo", "rld"]),
        ("hello (wo rld)", " ", None, ["hello", "(wo", "rld)"]),
    ],
)
def test_group_by_word(string, seperator, brackets, expected):
    assert expected == group_by_word(string, seperator=seperator, brackets=brackets)


@pytest.mark.parametrize(
    "string, seperator, brackets, error",
    [
        ("", " ", "(){}", ValueError),  # More than to brackets
        ("", " ", "((", ValueError),  # Non-unique brackets
        ("", "(", "()", ValueError),  # Seperator in brackets
        ("", "", "()", ValueError),  # No character seperator
        ("", None, "()", TypeError),  # Not valid seperator type
        ("", " ", ["(", ")"], TypeError),  # Not valid brackets type
        ("(hello", " ", "()", ValueError),  # No closing bracket
        ("(hello) (world", " ", "()", ValueError),  # No closing bracket
    ],
)
def test_group_by_word_errors(string, seperator, brackets, error):
    with pytest.raises(error):
        group_by_word(string, seperator=seperator, brackets=brackets)


@pytest.mark.parametrize(
    "variable, expected",
    [
        ("name", True),
        ("var_name", True),
        ("var_name12", True),
        ("Var_nAme12", True),
        ("var name", False),
        ("1name", False),
        ("var-name", False),
        ("var_name!", False),
    ],
)
def test_is_variable_name(variable, expected):
    assert is_variable_name(variable) == expected


@pytest.mark.parametrize(
    "number, expected",
    [
        ("1", True),
        ("-1", True),
        ("0123", True),
        ("0123456789", True),
        ("321891204189", True),
        ("-321891204189", True),
        ("a12", False),
        ("o", False),
        ("O", False),
        ("!", False),
    ],
)
def test_is_number(number, expected):
    assert is_number(number) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        ("1.", True),
        (".1", True),
        (".", False),
        ("1", False),
        ("01.23", True),
        ("01234.56789", True),
        ("32.1891204189", True),
        ("a12.", False),
        (".a12.", False),
        ("o.", False),
        ("O.", False),
        ("!.", False),
        (".o", False),
        (".O", False),
        (".!", False),
    ],
)
def test_is_float(value, expected):
    assert is_float(value) == expected



from __future__ import annotations

from netqasm.runtime.settings import Simulator, get_simulator

if get_simulator() == Simulator.NETSQUID:
    import netsquid as ns
    import numpy as np
    from netsquid.qubits import operators, qubitapi
    from netsquid.qubits.qubit import Qubit as NetSquidQubit


def qubit_from(phi: float, theta: float) -> NetSquidQubit:
    """Only used for simulation output purposes.
    Uses the phi and theta angles to construct a NetSquid qubit."""
    if get_simulator() != Simulator.NETSQUID:
        raise RuntimeError(
            "`qubit_from` function only possible with NetSquid simulator"
        )

    q = ns.qubits.create_qubits(1)[0]
    rot_y = operators.create_rotation_op(theta, (0, 1, 0))
    rot_z = operators.create_rotation_op(phi, (0, 0, 1))
    ns.qubits.operate(q, rot_y)
    ns.qubits.operate(q, rot_z)
    return q


def to_dm(q: NetSquidQubit) -> np.ndarray:
    """Only used for simulation output purposes."""
    if get_simulator() != Simulator.NETSQUID:
        raise RuntimeError("`to_dm` function only possible with NetSquid simulator")

    return ns.qubits.reduced_dm(q)  # type: ignore


def get_fidelity(q1: NetSquidQubit, q2: np.ndarray) -> float:
    """Only used for simulation output purposes.
    Gets the fidelity between the states q1 and q2"""
    if get_simulator() != Simulator.NETSQUID:
        raise RuntimeError(
            "`get_fidelity` function only possible with NetSquid simulator"
        )

    return qubitapi.fidelity(q1, q2)  # type: ignore



from netqasm.logging.glob import get_netqasm_logger
from netqasm.logging.output import get_new_app_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket
from netqasm.sdk.toolbox.measurements import parity_meas

logger = get_netqasm_logger()


def _get_default_strategy():
    return [
        ["XI", "-XZ", "IZ"],  # col 0
        ["XX", "YY", "ZZ"],  # col 1
        ["IX", "-ZX", "ZI"],  # col 2
    ]


def main(app_config=None, col=0, strategy=None):
    # This socket is only for post-processing purposes and not needed for the strategy to work.
    socket = Socket("player2", "player1", log_config=app_config.log_config)

    app_logger = get_new_app_logger(
        app_name=app_config.app_name, log_config=app_config.log_config
    )

    # Get the strategy
    if strategy is None:
        strategy = _get_default_strategy()
    if col >= len(strategy):
        raise ValueError(f"Not a col in the square {col}")

    # Create a EPR socket for entanglement generation
    epr_socket = EPRSocket("player1")

    # Initialize the connection
    player2 = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )
    with player2:
        # Create EPR pairs
        q1 = epr_socket.recv_keep()[0]
        q2 = epr_socket.recv_keep()[0]

        player2.flush()

        socket.recv_silent()

        # Make sure we order the qubits consistently with Player1
        # Get entanglement IDs
        q1_ID = q1.entanglement_info.sequence_number
        q2_ID = q2.entanglement_info.sequence_number

        if int(q1_ID) < int(q2_ID):
            qb = q1
            qd = q2
        else:
            qb = q2
            qd = q1

        # Wait for player 1 to finish their measurements
        socket.recv_silent()

        # Perform the three measurements
        app_logger.log(f"Measuring {strategy[col][0]} ...")
        m0 = parity_meas([qb, qd], strategy[col][0])
        player2.flush()
        app_logger.log(f"Outcome: {m0}")

        app_logger.log(f"Measuring {strategy[col][1]} ...")
        m1 = parity_meas([qb, qd], strategy[col][1])
        player2.flush()
        app_logger.log(f"Outcome: {m1}")

        app_logger.log(f"Measuring {strategy[col][2]} ...")
        m2 = parity_meas([qb, qd], strategy[col][2])
        player2.flush()
        app_logger.log(f"Outcome: {m2}")

    to_print = "\n\n"
    to_print += "==========================\n"
    to_print += "App player2: column is:\n"
    to_print += "(" + "_" * col + str(m0) + "_" * (2 - col) + ")\n"
    to_print += "(" + "_" * col + str(m1) + "_" * (2 - col) + ")\n"
    to_print += "(" + "_" * col + str(m2) + "_" * (2 - col) + ")\n"
    to_print += "==========================\n"
    to_print += "\n\n"
    logger.info(to_print)

    # Only needed for visualization: to check at which cell the column intersects with the row of the other player.
    player1_row = eval(socket.recv_silent())
    player1_outcomes = eval(socket.recv_silent())
    col_outcomes = [int(m0), int(m1), int(m2)]

    square = [["", "", ""], ["", "", ""], ["", "", ""]]

    square[player1_row][0] = str(player1_outcomes[0])
    square[player1_row][1] = str(player1_outcomes[1])
    square[player1_row][2] = str(player1_outcomes[2])
    square[0][col] = str(col_outcomes[0])
    square[1][col] = str(col_outcomes[1])
    square[2][col] = str(col_outcomes[2])
    square[player1_row][col] = f"{player1_outcomes[col]}/{col_outcomes[player1_row]}"

    # table = []
    # for row in range(3):
    #     table.append(
    #         square[row]
    #     )

    return {
        "col": col_outcomes,
        "square": square,
    }


if __name__ == "__main__":
    main()



"""Abstractions for the classical memory shared between the Host and
the quantum node controller.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union

from netqasm.lang import operand
from netqasm.lang.encoding import ADDRESS_BITS, REG_INDEX_BITS, RegisterName
from netqasm.lang.ir import Symbols
from netqasm.lang.parsing import parse_address, parse_register
from netqasm.runtime.settings import get_is_using_hardware


def _assert_within_width(value: int, width: int) -> None:
    if not get_is_using_hardware():
        # in simulation, don't care about overflow
        return
    min_value = -(2 ** (width - 1))
    max_value = 2 ** (width - 1) - 1
    if not min_value <= value <= max_value:
        raise OverflowError(f"value {value} does not fit into {width} bits")


class RegisterGroup:
    """A register group (like "R", or "Q") in shared memory."""

    def __init__(self):
        self._size: int = 2**REG_INDEX_BITS
        self._register: Dict[int, Optional[int]] = {}

    def __len__(self) -> int:
        return self._size

    def __str__(self) -> str:
        return str(self._register)

    def __setitem__(self, index: int, value: int) -> None:
        self._assert_within_length(index)
        _assert_within_width(value, ADDRESS_BITS)
        self._register[index] = value

    def __getitem__(self, index: int) -> Optional[int]:
        self._assert_within_length(index)
        return self._register.get(index)

    def _assert_within_length(self, index: int) -> None:
        if not (0 <= index < len(self)):
            raise IndexError(f"index {index} is not within 0 and {len(self)}")

    def _get_active_values(self) -> List[Tuple[int, int]]:
        return [
            (index, value)
            for index, value in self._register.items()
            if value is not None
        ]


def setup_registers() -> Dict[RegisterName, RegisterGroup]:
    return {reg_name: RegisterGroup() for reg_name in RegisterName}


class Arrays:
    def __init__(self):
        self._arrays: Dict[int, List[Optional[int]]] = {}

    # TODO add test for this
    def _get_active_values(self) -> List[Tuple[operand.ArrayEntry, int]]:
        values = []
        for address, array in self._arrays.items():
            for index, value in enumerate(array):
                if value is None:
                    continue
                address_entry = parse_address(
                    f"{Symbols.ADDRESS_START}{address}"
                    f"{Symbols.INDEX_BRACKETS[0]}{index}{Symbols.INDEX_BRACKETS[1]}"
                )
                if not isinstance(address_entry, operand.ArrayEntry):
                    raise RuntimeError(
                        f"Something went wrong: address_entry should be "
                        f"ArrayEntry but it is {type(address_entry)}"
                    )
                values.append((address_entry, value))
        return values

    def __str__(self) -> str:
        return str(self._arrays)

    def __setitem__(
        self,
        key: Tuple[int, Union[int, slice]],
        value: Union[None, int, List[Optional[int]]],
    ) -> None:
        address, index = self._extract_key(key)
        if isinstance(index, int):
            if isinstance(value, int):
                _assert_within_width(value, ADDRESS_BITS)
                _assert_within_width(index, ADDRESS_BITS)
        elif isinstance(index, slice):
            self._assert_list(value)
            if index.start is not None:
                _assert_within_width(index.start, ADDRESS_BITS)
            if index.stop is not None:
                _assert_within_width(index.stop, ADDRESS_BITS)
        else:
            raise TypeError(f"Cannot use {key} of type {type(key)} as an index")
        array = self._get_array(address)
        try:
            if isinstance(index, slice):
                assert isinstance(value, list)
                assert len(array[index]) == len(value), "value not of correct length"
            array[index] = value  # type: ignore
        except IndexError:
            raise IndexError(
                f"index {index} is out of range for array with address {address}"
            )

    def __getitem__(
        self, key: Tuple[int, Union[int, slice]]
    ) -> Union[None, int, List[Optional[int]]]:
        address, index = self._extract_key(key)
        try:
            array = self._get_array(address)
        except IndexError:
            return None

        try:
            value = array[index]
        except IndexError:
            raise IndexError(
                f"index {index} is out of range for array with address {address}"
            )
        return value

    def _get_array(self, address: int) -> List[Optional[int]]:
        if address not in self._arrays:
            raise IndexError(f"No array with address {address}")
        return self._arrays[address]

    def _set_array(self, address: int, array: List[Optional[int]]) -> None:
        if address not in self._arrays:
            raise IndexError(f"No array with address {address}")
        self._assert_list(array)
        self._arrays[address] = array

    def has_array(self, address: int) -> bool:
        return address in self._arrays

    @staticmethod
    def _extract_key(
        key: Tuple[int, Union[int, slice]]
    ) -> Tuple[int, Union[int, slice]]:
        try:
            address, index = key
        except (TypeError, ValueError):
            raise ValueError(
                "Can only access entries and slices of arrays, not the full array"
            )
        _assert_within_width(address, ADDRESS_BITS)
        return address, index

    @staticmethod
    def _assert_list(value: Any) -> None:
        if not isinstance(value, list):
            raise TypeError(f"expected 'list', not {type(value)}")
        for x in value:
            if x is not None:
                _assert_within_width(x, ADDRESS_BITS)
        _assert_within_width(len(value), ADDRESS_BITS)

    def init_new_array(self, address: int, length: int) -> None:
        # TODO, is it okay to overwrite the array if it exists?
        _assert_within_width(address, ADDRESS_BITS)
        self._arrays[address] = [None] * length


class SharedMemory:
    """Representation of the classical memory that is shared between the Host
    and the quantum node controller.

    Each application is associated with a single `SharedMemory`.
    Although the name contains "shared", it is typically only used in one "direction":
    the quantum node controller writes to it, and the Host reads from it.

    Shared memory consists of two types of memory: registers and arrays.
    The quantum node controller writes to the shared memory if it executes a "return"
    NetQASM instruction (`ret_reg` or `ret_arr`). This is typically done at the end
    of a subroutine. The Host can then read the values from the shared memory, which
    means that the Host effectively receives the returned values.

    When trying to use a value represented by a `Future`, it will try to get the value
    from the shared memory. So, *only* after the quantum node controller has executed
    a subroutine containing the relevant `ret_reg` or `ret_arr` NetQASM instruction,
    the shared memory is updated and the value from the Future can be used.

    The specific runtime context determines how a shared memory is implemented and
    by which component it is controlled. In the case of a physical setup, where the
    Host and the quantum node controller may be separate devices, "writing to the
    shared memory" may simply be "sending values from the quantum node controller to
    the Host". On the other hand, a simulator that runs the Host and the quantum node
    controller in the same process might e.g. use a global shared object.
    """

    def __init__(self):
        self._registers: Dict[RegisterName, RegisterGroup] = setup_registers()
        self._arrays: Arrays = Arrays()

    def __getitem__(
        self, key: Union[operand.Register, Tuple[int, Union[int, slice]], int]
    ) -> Union[None, int, List[Optional[int]]]:
        if isinstance(key, operand.Register):
            return self.get_register(key)
        elif isinstance(key, tuple):
            address, index = key
            return self.get_array_part(address, index)
        elif isinstance(key, int):
            return self._get_array(key)

    def get_register(self, register: Union[str, operand.Register]) -> Optional[int]:
        if isinstance(register, str):
            reg = parse_register(register)
        else:
            reg = register
        return self._registers[reg.name][reg.index]

    def set_register(self, register: Union[str, operand.Register], value: int) -> None:
        if isinstance(register, str):
            reg = parse_register(register)
        else:
            reg = register
        self._registers[reg.name][reg.index] = value

    def get_array_part(
        self, address: int, index: Union[int, slice]
    ) -> Union[None, int, List[Optional[int]]]:
        return self._arrays[address, index]

    def set_array_part(
        self,
        address: int,
        index: Union[int, slice],
        value: Union[None, int, List[Optional[int]]],
    ):
        self._arrays[address, index] = value

    def _get_array(self, address: int) -> List[Optional[int]]:
        return self._arrays._get_array(address)

    def init_new_array(
        self,
        address: int,
        length: int = 1,
        new_array: Optional[List[Optional[int]]] = None,
    ) -> None:
        if new_array is not None:
            length = len(new_array)
        self._arrays.init_new_array(address, length)
        if new_array is not None:
            self._arrays._set_array(address, new_array)

    def _get_active_values(
        self,
    ) -> List[Union[Tuple[operand.Register, int], Tuple[operand.ArrayEntry, int]]]:
        all_values: List[
            Union[Tuple[operand.Register, int], Tuple[operand.ArrayEntry, int]]
        ] = []
        for reg_name, reg in self._registers.items():
            act_reg_values = reg._get_active_values()
            reg_values = [
                (parse_register(f"{reg_name.name}{index}"), value)
                for index, value in act_reg_values
            ]
            all_values += reg_values
        all_values += self._arrays._get_active_values()
        return all_values


class SharedMemoryManager:
    """Global object that manages shared memories. Typically used by simulators.

    This class can be used as a global object that stores `SharedMemory` objects.
    Simulators that simulate Hosts and the quantum node controllers in the same
    process may use this to have a single location for creating and accessing shared
    memories.
    """

    _MEMORIES: Dict[Tuple[str, Optional[int]], Optional[SharedMemory]] = {}

    @classmethod
    def create_shared_memory(
        cls, node_name: str, key: Optional[int] = None
    ) -> SharedMemory:
        absolute_key = (node_name, key)
        if cls._MEMORIES.get(absolute_key) is not None:
            raise RuntimeError(
                f"Shared memory for (node, key): ({node_name}, {key}) already exists."
            )
        memory = SharedMemory()
        cls._MEMORIES[absolute_key] = memory
        return memory

    @classmethod
    def get_shared_memory(
        cls, node_name: str, key: Optional[int] = None
    ) -> Optional[SharedMemory]:
        absolute_key = (node_name, key)
        memory = cls._MEMORIES.get(absolute_key)
        return memory

    @classmethod
    def reset_memories(cls) -> None:
        for key in list(cls._MEMORIES.keys()):
            cls._MEMORIES.pop(key)



from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

QubitState = Tuple[Tuple[complex, complex], Tuple[complex, complex]]  # 2x2 matrix
AbsoluteQubitID = List[Union[str, int]]  # [app_name, qubit_id]


@dataclass
class QubitGroup:
    is_entangled: Optional[bool]
    qubit_ids: List[AbsoluteQubitID]
    state: Optional[QubitState]  # only set when group size is 1


QubitGroups = Dict[int, QubitGroup]  # group_id -> qubit_group


class EntanglementType(Enum):
    CK = "CK"  # Create and Keep
    MD = "MD"  # Measure Directly


class EntanglementStage(Enum):
    START = "start"
    FINISH = "finish"


@dataclass
class InstrLogEntry:
    WCT: str
    """Wall clock time. Format is Python's `datetime.now()`."""

    SIT: int
    """ Time in NetSquid simulation, in nanoseconds."""

    AID: int
    """ App ID, used internally by the backend."""

    SID: int
    """ Subroutine ID. Used internally by the Executor."""

    PRC: int
    """ Program counter. Used internally by the Executor."""

    HLN: int
    """
    Host line number.
    Line number in source file (.py) related to what is currently executed.
    The line is in the file given by HFL (see below).
    """

    HFL: str
    """
    Host file.
    Source file (.py) of current executed instruction.
    """

    INS: str
    """ Mnemonic of the NetQASM instruction being executed."""

    OPR: List[str]
    """
    Operands (register, array-entries..).
    List of "op=val" strings
    """

    ANG: Optional[Dict[str, int]]
    """
    Angle represented as the fraction num/den.
    For non-rotation instructions, ANG is None.
    For rotation instructions ANG is a dictionary with 2 entries:
    'num' (an int) and 'den' (an int).
    """

    QID: List[int]
    """ Physical qubit IDs of qubits part of the current operation."""

    VID: List[int]
    """ Virtual qubit IDs of qubits part of the current operation."""

    OUT: Optional[int]
    """ Measurement outcome. Only set in case of a measurement instruction."""

    QGR: Optional[QubitGroups]
    """ Dictionary specifying groups of qubits across the network."""

    LOG: str
    """ Human-readable message."""


@dataclass
class NetworkLogEntry:
    WCT: str
    """ Wall clock time. Format is Python's `datetime.now()`."""

    SIT: int
    """ Time in NetSquid simulation, in nanoseconds."""

    TYP: Optional[EntanglementType]
    """
    Entanglement generation type(Measure Directly or Create and Keep).
    For the 'start' entanglement stage (see INS below), this value is None
    since at this stage the value cannot be determined yet.
    For the 'finish' stage, the correct value is filled in, however.
    """

    INS: EntanglementStage
    """ Entanglement generation stage(start or finish)."""

    BAS: Optional[List[int]]
    """
    Bases in which the two qubits(one on each end) were measured in .
    Only applies to the Measure Directly case. It is `None` otherwise.
    """

    MSR: List[int]
    """
    Measurement outcomes of the two qubits(one on each end).
    Only applies to the Measure Directly case. It is `None` otherwise.
    """

    NOD: List[str]
    """ Node names involved in this entanglement operation."""

    PTH: List[str]
    """
    Path of links used for entanglement generation.
    Links are identified using their names.
    """

    QID: List[int]
    """ Physical qubit IDs of qubits part of the current operation."""

    QGR: Optional[QubitGroups]
    """
    Dictionary specifying groups of qubits across the network,
    as they are after the current operation.
    """

    LOG: str
    """ Human-readable message."""


@dataclass
class ClassCommLogEntry:
    WCT: str
    """ Wall clock time. Format is Python's `datetime.now()`."""

    HLN: int
    """
    Host line number.
    Line number in source file (.py) related to what is currently executed.
    The line is in the file given by HFL(see below).
    """

    HFL: str
    """
    Host file.
    Source file (.py) of current executed instruction.
    """

    INS: str
    """ Classical operation being performed."""

    MSG: str
    """ Message that is sent or received."""

    SEN: str
    """ Name(role) of the sender."""

    REC: str
    """ Name(role) of the receiver."""

    SOD: str
    """ Socket ID(used internally)."""

    LOG: str
    """ Human-readable message."""


@dataclass
class AppLogEntry:
    WCT: str
    """ Wall clock time. Format is Python's `datetime.now()`."""

    HLN: int
    """
    Host line number.
    Line number in source file (.py) related to what is currently executed.
    The line is in the file given by HFL(see below).
    """

    # Source file (.py) of current executed instruction.
    """ Host file."""
    HFL: str

    LOG: str
    """ Human-readable message."""



import os
from enum import Enum


class Simulator(Enum):
    NETSQUID = "netsquid"
    NETSQUID_SINGLE_THREAD = "netsquid_single_thread"
    SIMULAQRON = "simulaqron"
    DEBUG = "debug"


class Formalism(Enum):
    STAB = "stab"
    KET = "ket"
    DM = "dm"


class Flavour(Enum):
    VANILLA = "vanilla"
    NV = "nv"


SIMULATOR_ENV = "NETQASM_SIMULATOR"


def set_simulator(simulator):
    os.environ[SIMULATOR_ENV] = Simulator(simulator).value


def get_simulator():
    simulator = os.environ.get(SIMULATOR_ENV)
    if simulator is None:
        return _default_simulator()
    else:
        return Simulator(simulator)


def _default_simulator():
    return Simulator.NETSQUID


_is_using_hardware = False


def set_is_using_hardware(b):
    global _is_using_hardware
    _is_using_hardware = b


def get_is_using_hardware():
    return _is_using_hardware



from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection
from netqasm.sdk.toolbox import set_qubit_state


def main(app_config=None, phi=0.0, theta=0.0):
    log_config = app_config.log_config
    # Initialize the connection to the backend
    alice = NetQASMConnection(
        app_name="alice",
        log_config=log_config,
    )
    with alice:
        q = Qubit(alice)
        set_qubit_state(q, phi, theta)

        q2 = Qubit(alice)
        q.cnot(q2)
        q.H()
        m1 = q.measure()
        m2 = q2.measure()

    # Send the correction information
    m1, m2 = int(m1), int(m2)
    print(f"m1 = {m1}, m2 = {m2}")

    return {"m1": m1, "m2": m2}


if __name__ == "__main__":
    main()



from dataclasses import dataclass

import numpy as np

from netqasm.util.quantum_gates import (
    get_controlled_rotation_matrix,
    get_rotation_matrix,
)

from . import core

# Explicit instruction types in the NV flavour.


@dataclass
class RotXInstruction(core.RotationInstruction):
    id: int = 27
    mnemonic: str = "rot_x"

    def to_matrix(self) -> np.ndarray:
        axis = [1, 0, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class RotYInstruction(core.RotationInstruction):
    id: int = 28
    mnemonic: str = "rot_y"

    def to_matrix(self) -> np.ndarray:
        axis = [0, 1, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class RotZInstruction(core.RotationInstruction):
    id: int = 29
    mnemonic: str = "rot_z"

    def to_matrix(self) -> np.ndarray:
        axis = [0, 0, 1]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class ControlledRotXInstruction(core.ControlledRotationInstruction):
    id: int = 30
    mnemonic: str = "crot_x"

    def to_matrix(self) -> np.ndarray:
        axis = [1, 0, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_controlled_rotation_matrix(axis, angle)

    def to_matrix_target_only(self) -> np.ndarray:
        axis = [1, 0, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class ControlledRotYInstruction(core.ControlledRotationInstruction):
    id: int = 31
    mnemonic: str = "crot_y"

    def to_matrix(self) -> np.ndarray:
        axis = [1, 0, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_controlled_rotation_matrix(axis, angle)

    def to_matrix_target_only(self) -> np.ndarray:
        axis = [1, 0, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)



import logging

import pytest

from netqasm.backend.executor import Executor
from netqasm.lang.encoding import RegisterName
from netqasm.lang.operand import Register
from netqasm.lang.parsing import parse_text_subroutine
from netqasm.logging.glob import set_log_level
from netqasm.sdk.shared_memory import SharedMemoryManager


@pytest.mark.parametrize(
    "subroutine_str, expected_register, expected_output",
    [
        (
            """
        # NETQASM 1.0
        # APPID 0
        # DEFINE op h
        # DEFINE q Q0
        # DEFINE m M0
        set $q 0
        qalloc $q
        init $q
        $op $q // this is a comment
        meas $q $m
        // this is also a comment
        beq $m 0 EXIT
        x $q
        EXIT:
        qfree $q
        """,
            Register(RegisterName.M, 0),
            0,
        ),
        (
            """
        # NETQASM 1.0
        # APPID 0
        # DEFINE i R0
        set $i 0
        LOOP:
        beq $i 10 EXIT
        add $i $i 1
        beq 0 0 LOOP
        EXIT:
        """,
            Register(RegisterName.R, 0),
            10,
        ),
    ],
)
def test_executor(subroutine_str, expected_register, expected_output):
    set_log_level(logging.DEBUG)
    subroutine = parse_text_subroutine(subroutine_str)

    print(subroutine)

    SharedMemoryManager.reset_memories()

    app_id = 0
    executor = Executor()
    # Consume the generator
    executor.init_new_application(app_id=app_id, max_qubits=1)
    for _ in range(10):
        list(executor.execute_subroutine(subroutine=subroutine))
        assert executor._get_register(app_id, expected_register) == expected_output


@pytest.mark.parametrize(
    "subroutine_str, error_type, error_line",
    [
        (
            """
        # NETQASM 0.0
        # APPID 0
        set R0 1
        add R0 R0 R0
        set R1 0
        addm R0 R0 R0 R1
        """,
            RuntimeError,
            3,
        ),
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        qalloc Q0
        qalloc Q0
        """,
            RuntimeError,
            2,
        ),
    ],
)
def test_failing_executor(subroutine_str, error_type, error_line):
    set_log_level(logging.DEBUG)
    subroutine = parse_text_subroutine(subroutine_str)

    print(subroutine)

    SharedMemoryManager.reset_memories()

    app_id = 0
    executor = Executor()
    executor.init_new_application(app_id=app_id, max_qubits=1)

    with pytest.raises(error_type) as exc:
        executor.consume_execute_subroutine(subroutine=subroutine)

    print(f"Exception: {exc.value}")
    assert str(exc.value).startswith(f"At line {error_line}")


if __name__ == "__main__":
    subroutine_str = """
        # NETQASM 1.0
        # APPID 0
        # DEFINE i R0
        set $i 0
        LOOP:
        beq $i 10 EXIT
        add $i $i 1
        beq 0 0 LOOP
        EXIT:
        """
    expected_register = Register(RegisterName.R, 0)
    expected_output = 10

    test_executor(subroutine_str, expected_register, expected_output)

    subroutine_str = """
        # NETQASM 0.0
        # APPID 0
        set R0 1
        add R0 R0 R0
        set R1 0
        addm R0 R0 R0 R1
        """
    error_type = RuntimeError
    error_line = 3

    test_failing_executor(subroutine_str, error_type, error_line)



import numpy as np
import pytest

from netqasm.lang.instr.core import LoadInstruction, SetInstruction
from netqasm.lang.operand import Immediate
from netqasm.lang.subroutine import Subroutine
from netqasm.logging.glob import get_netqasm_logger
from netqasm.qlink_compat import BellState
from netqasm.runtime.application import default_app_instance
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def simulate_different_bell_state(
    subrt: Subroutine, bell_state: BellState
) -> Subroutine:
    # Change "Load bell state index from array" instruction into a simple "set"
    # It is always the second 'load' instruction in the subroutine.
    found_first_load = False
    instr_index = None
    for i, instr in enumerate(subrt.instructions):
        if isinstance(instr, LoadInstruction):
            if not found_first_load:
                found_first_load = True
            else:
                instr_index = i
                break

    assert instr_index is not None

    orig_instr: LoadInstruction = subrt.instructions[instr_index]  # type: ignore
    subrt.instructions[instr_index] = SetInstruction(
        reg=orig_instr.reg, imm=Immediate(bell_state.value)
    )
    return subrt


def post_function(backend):
    alice_state = backend.nodes["Alice"].qmemory._get_qubits(0)[0].qstate
    bob_state = backend.nodes["Bob"].qmemory._get_qubits(0)[0].qstate
    assert alice_state is bob_state
    expected_state = np.array(
        [[0.5, 0, 0, 0.5], [0, 0, 0, 0], [0, 0, 0, 0], [0.5, 0, 0, 0.5]]
    )

    logger.info(f"state = {alice_state.qrepr.reduced_dm()}")
    assert np.all(np.isclose(expected_state, alice_state.qrepr.reduced_dm()))


def run_alice():
    epr_socket = EPRSocket("Bob")
    with NetQASMConnection("Alice", epr_sockets=[epr_socket]):
        # Create entanglement
        epr_socket.create_keep()[0]


def run_bob_phi_plus():
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]):
        epr_socket.recv_keep()


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_create_epr_phi_plus():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob_phi_plus),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )


def run_bob_phi_minus():
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]) as conn:
        [q] = epr_socket.recv_keep()
        subrt = conn.compile()
        # pretend PHI_MINUS was generated
        subrt2 = simulate_different_bell_state(subrt, BellState.PHI_MINUS)
        conn.commit_subroutine(subrt2)
        # compiler should apply a Z, undo it:
        q.Z()


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_create_epr_phi_minus():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob_phi_minus),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )


def run_bob_psi_plus():
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]) as conn:
        [q] = epr_socket.recv_keep()
        subrt = conn.compile()
        # pretend PSI_PLUS was generated
        subrt2 = simulate_different_bell_state(subrt, BellState.PSI_PLUS)
        conn.commit_subroutine(subrt2)
        # compiler should apply an X, undo it:
        q.X()


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_create_epr_psi_plus():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob_psi_plus),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )


def run_bob_psi_minus():
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]) as conn:
        [q] = epr_socket.recv_keep()
        subrt = conn.compile()
        # pretend PSI_MINUS was generated
        subrt2 = simulate_different_bell_state(subrt, BellState.PSI_MINUS)
        conn.commit_subroutine(subrt2)
        # compiler should apply an X and a Z, undo it:
        q.Z()
        q.X()


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_create_epr_psi_minus():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob_psi_minus),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )


if __name__ == "__main__":
    test_create_epr_phi_plus()
    test_create_epr_phi_minus()
    test_create_epr_psi_plus()
    test_create_epr_psi_minus()



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    with NetQASMConnection("Alice") as alice:
        num = 100

        outcomes = alice.new_array(num)

        def body(alice, _):
            q = Qubit(alice)
            q.H()
            q.measure(future=outcomes.get_future_index("R0"))

        alice.loop_body(body, stop=num, loop_register="R0")
        alice.flush()
        assert len(outcomes) == num
        avg = sum(outcomes) / num
        logger.info(f"Average: {avg}")
        assert 0.4 <= avg <= 0.6


def test_measure_loop():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



from netqasm.logging.glob import set_log_level
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.epr_socket import EPRSocket

DebugConnection.node_ids = {
    "Alice": 0,
    "Bob": 1,
}


def main(no_output=False):
    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        m = epr_socket.create_rsp()[0]
        print(m.raw_measurement_outcome)

    if no_output:
        # Third message since the first two are for init application
        # and opening the EPR socket
        print(f"binary:\n{alice.storage[2]}")


if __name__ == "__main__":
    set_log_level("INFO")
    main()



import numpy as np

from netqasm.examples.lib.bqc import (
    measXY,
    recv_meas_cmd,
    recv_teleported_state,
    send_meas_outcome,
)
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()


def main(app_config=None):
    socket = Socket("server", "client", log_config=app_config.log_config)
    epr_socket = EPRSocket("client")

    num_qubits = 4

    server = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
        max_qubits=num_qubits,
    )

    with server:
        # Receive qubits q0 to q3 from the Client by teleportation.
        q = [None] * num_qubits
        for i in range(num_qubits):
            q[i] = recv_teleported_state(epr_socket)

        # Apply CPHASE gates between neighbouring nodes.
        # (See cluster state in the README.)
        for i in range(num_qubits - 1):
            q[i].cphase(q[i + 1])
        q[0].cphase(q[2])

        # TODO check why this is needed
        server.flush()

        # Receive from the Client the angle to measure q1 in.
        delta1 = recv_meas_cmd(socket)
        s1 = measXY(q[1], delta1)
        server.flush()
        send_meas_outcome(socket, s1)

        # Receive from the Client the angle to measure q2 in.
        delta2 = recv_meas_cmd(socket)
        s2 = measXY(q[2], delta2)
        server.flush()
        send_meas_outcome(socket, s2)

        # Measure the output qubits (q0 and q3) in the Y-basis.
        m0 = measXY(q[0], np.pi / 2)
        m1 = measXY(q[3], np.pi / 2)
        server.flush()

        # Send the measurement outcomes to the Client.
        send_meas_outcome(socket, m0)
        send_meas_outcome(socket, m1)
